from __future__ import absolute_import

from gutter.django import registry

from gutter.client import arguments


class User(arguments.Container):
    name = arguments.String(lambda self: 'Jeff')
    age = arguments.Value(lambda self: 29)
    registered_on = arguments.Boolean(lambda self: True)


class Request(arguments.Container):
    ip = arguments.String('192.168.0.1')


registry.arguments.register(User.name)
registry.arguments.register(User.age)
registry.arguments.register(User.registered_on)
registry.arguments.register(Request.ip)



from gutter.client.operators.comparable import *
from gutter.client.operators.identity import *
from gutter.client.operators.misc import *

from itertools import groupby

from operator import attrgetter, itemgetter


class OperatorsDict(dict):

    def __init__(self, *operators):
        map(self.register, operators)

    def register(self, operator):
        self[operator.name] = operator

    @property
    def as_choices(self):
        groups = {}

        for operator in sorted(self.values(), key=attrgetter('preposition')):
            key = operator.group.title()
            pair = (operator.name, operator.preposition.title())

            groups.setdefault(key, [])
            groups[key].append(pair)

        return sorted(groups.items(), key=itemgetter(0))

    @property
    def arguments(self):
        return dict((name, op.arguments) for name, op in self.items())


class ArgumentsDict(dict):

    @property
    def as_choices(self):
        sorted_strings = sorted(map(str, self.values()))
        extract_classname = lambda a: a.split('.')[0]

        grouped = groupby(sorted_strings, extract_classname)

        groups = {}
        for name, arguments in grouped:
            groups.setdefault(name, [])
            groups[name].extend((a, a) for a in arguments)

        return groups.items()

    def register(self, argument):
        self[str(argument)] = argument


operators = OperatorsDict(
    Equals,
    Between,
    LessThan,
    LessThanOrEqualTo,
    MoreThan,
    MoreThanOrEqualTo,
    Truthy,
    Percent,
    PercentRange
)

arguments = ArgumentsDict()



from django import forms
from django.core.validators import RegexValidator
from django.forms.formsets import formset_factory, BaseFormSet
from django.forms.widgets import Select, Textarea
from django.utils.html import escape, conditional_escape
from django.utils.encoding import force_unicode

from itertools import chain

from gutter.django.registry import operators, arguments
from gutter.client.models import Switch, Condition

from functools import partial


class OperatorSelectWidget(Select):

    def __init__(self, arguments, *args, **kwargs):
        self.arguments = arguments
        super(OperatorSelectWidget, self).__init__(*args, **kwargs)

    def render_options(self, choices, selected_choices):
        def render_option(option_value, option_label):
            option_value = force_unicode(option_value)
            selected_html = (option_value in selected_choices) and u' selected="selected"' or ''
            return u'<option data-arguments="%s" value="%s"%s>%s</option>' % (
                ','.join(self.arguments[option_value]),
                escape(option_value), selected_html,
                conditional_escape(force_unicode(option_label)))

        # Normalize to strings.
        selected_choices = set([force_unicode(v) for v in selected_choices])
        output = []

        for option_value, option_label in chain(self.choices, choices):
            if isinstance(option_label, (list, tuple)):
                output.append(u'<optgroup label="%s">' % escape(force_unicode(option_value)))
                for option in option_label:
                    output.append(render_option(*option))
                output.append(u'</optgroup>')
            else:
                output.append(render_option(option_value, option_label))
        return u'\n'.join(output)


class SwitchForm(forms.Form):

    STATES = {1: 'Disabled', 2: 'Selective', 3: 'Global'}.items()
    SWITCH_NAME_REGEX_VALIDATOR = RegexValidator(
        regex=r'^[\w_:]+$',
        message='Must only be alphanumeric, underscore, and colon characters.'
    )

    name = forms.CharField(max_length=100)
    label = forms.CharField(required=False)
    description = forms.CharField(widget=Textarea(), required=False)
    state = forms.IntegerField(widget=Select(choices=STATES))

    compounded = forms.BooleanField(required=False)
    concent = forms.BooleanField(required=False)

    delete = forms.BooleanField(required=False)

    name.validators.append(SWITCH_NAME_REGEX_VALIDATOR)

    @classmethod
    def from_object(cls, switch):
        data = dict(
            label=switch.label,
            name=switch.name,
            description=switch.description,
            state=switch.state,
            compounded=switch.compounded,
            concent=switch.concent
        )

        instance = cls(initial=data)

        condition_dicts = map(ConditionForm.to_dict, switch.conditions)
        instance.conditions = ConditionFormSet(initial=condition_dicts)
        instance.fields['name'].widget.attrs['readonly'] = True

        return instance

    def field(self, key):
        return self.data.get(key, None) or self.initial[key]

    @property
    def to_object(self):
        switch = Switch(
            name=self.cleaned_data['name'],
            label=self.cleaned_data['label'],
            description=self.cleaned_data['description'],
            state=self.cleaned_data['state'],
            compounded=self.cleaned_data['compounded'],
            concent=self.cleaned_data['concent'],
        )

        return switch


class ConditionForm(forms.Form):

    negative_widget = Select(choices=((False, 'Is'), (True, 'Is Not')))

    argument = forms.ChoiceField(choices=arguments.as_choices)
    negative = forms.BooleanField(widget=negative_widget, required=False)
    operator = forms.ChoiceField(
        choices=operators.as_choices,
        widget=OperatorSelectWidget(operators.arguments)
    )

    @staticmethod
    def to_dict(condition):
        fields = dict(
            argument='.'.join((condition.argument.__name__, condition.attribute)),
            negative=condition.negative,
            operator=condition.operator.name
        )

        fields.update(condition.operator.variables)

        return fields


class BaseConditionFormSet(BaseFormSet):

    @property
    def to_objects(self):
        return map(self.__make_condition, self.forms)

    def __make_condition(self, form):
        data = form.cleaned_data.copy()

        # Extract out the values from the POST data.  These are all strings at
        # this point
        operator_str = data.pop('operator')
        negative_str = data.pop('negative')
        argument_str = data.pop('argument')

        # Operators in the registry are the types (classes), so extract that out
        # and we will construct it from the remaining data, which are the
        # arguments for the operator
        operator_type = operators[operator_str]

        # Arguments are a Class property, so just a simple fetch from the
        # arguments dict will retreive it
        argument = arguments[argument_str]

        # The remaining variables in the data are the arguments to the operator,
        # but they need to be cast by the argument to their right type
        caster = argument.variable.to_python
        data = dict((k, caster(v)) for k, v in data.items())

        return Condition(
            argument=argument.owner,
            attribute=argument.name,
            operator=operator_type(**data),
            negative=bool(int(negative_str))
        )

    def value_at(self, index, field):
        if self.initial:
            return self.initial[index][field]
        elif index is not None:
            return self.data['form-%s-%s' % (index, field)]

    def add_fields(self, form, index):
        value = partial(self.value_at, index)

        for argument in operators.arguments.get(value('operator'), []):
            form.fields[argument] = forms.CharField(initial=value(argument))

        super(BaseConditionFormSet, self).add_fields(form, index)


ConditionFormSet = formset_factory(
    ConditionForm,
    formset=BaseConditionFormSet,
    extra=0
)


class SwitchFormManager(object):

    def __init__(self, switch, condition_set):
        self.switch = switch
        self.conditions = condition_set

    @classmethod
    def from_post(cls, post_data):
        return cls(SwitchForm(post_data), ConditionFormSet(post_data))

    def is_valid(self):
        return self.switch.is_valid() and self.conditions.is_valid()

    def save(self, gutter_manager):
        switch = self.switch.to_object
        switch.conditions = self.conditions.to_objects
        gutter_manager.register(switch)

    def add_to_switch_list(self, switches):
        self.switch.conditions = self.conditions
        switches.insert(0, self.switch)

    def delete(self, gutter_manager):
        gutter_manager.unregister(self.switch.data['name'])



import logging


logger = logging.getLogger(__name__)


def discover():
    """
    Auto-discover any Gutter configuration present in the django
    INSTALLED_APPS.
    """
    from django.conf import settings
    from django.utils.importlib import import_module

    for app in settings.INSTALLED_APPS:
        module = '%s.gutter' % app

        try:
            import_module(module)
            logger.info('Successfully autodiscovered %s' % module)
        except:
            pass



"""
gutter.nexus_modules
~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2010 DISQUS.
:license: Apache License 2.0, see LICENSE for more details.
"""

from __future__ import absolute_import

import os

from gutter.client.default import gutter as manager
from gutter.django.forms import SwitchForm, ConditionFormSet, SwitchFormManager
import nexus


def operator_info(operator):
    return dict(
        path=operator,
        label=operator.label,
        preposition=operator.preposition,
        arguments=operator.arguments,
        group=operator.group
    )


def argument_info(argument_container):
    return dict(
        string=argument_container.__name__,
        arguments=argument_container.arguments
    )


class GutterModule(nexus.NexusModule):
    home_url = 'index'
    name = 'gutter'
    media_root = os.path.normpath(os.path.join(os.path.dirname(__file__), 'media'))

    def get_title(self):
        return 'Gutter'

    def get_urls(self):
        from django.conf.urls.defaults import patterns, url

        urlpatterns = patterns('',
            url(r'^update/$', self.as_view(self.update), name='update'),
            url(r'^$', self.as_view(self.index), name='index'),
        )

        return urlpatterns

    def render_on_dashboard(self, request):
        return 'switches'

    @property
    def __index_context(self):
        switches = map(SwitchForm.from_object, manager.switches)
        switches = sorted(switches, key=lambda x: x.field('name'))

        new_switch = SwitchForm()
        new_switch.conditions = ConditionFormSet()

        return {
            "switches": switches,
            "new_switch": new_switch
        }

    def __render(self, request, invalid_manager=None, **notices):
        context = self.__index_context
        template = "gutter/index.html"

        context.update(notices=notices)

        if invalid_manager:
            invalid_manager.add_to_switch_list(context['switches'])
            context['notices']['error'] = 'Unable to save switch.'

        return self.render_to_response(template, context, request)

    def index(self, request):
        return self.__render(request)

    def update(self, request):
        form_manager = SwitchFormManager.from_post(request.POST)

        if form_manager.switch.data.get('delete'):
            print request.POST
            manager.unregister(form_manager.switch.data['name'])
            return self.__render(request, success='Switch deleted successfully.')
        elif form_manager.is_valid():
            form_manager.save(manager)
            return self.__render(request, success='Switch saved successfully.')
        else:
            return self.__render(request, invalid_manager=form_manager)


nexus.site.register(GutterModule, 'gutter')



#!/usr/bin/env python

import sys
from setuptools import find_packages


try:
    from notsetuptools import setup
except ImportError:
    from setuptools import setup

try:
    import multiprocessing
except ImportError:
    pass

tests_require = ['nose', 'unittest2', 'exam', 'redis', 'durabledict']

dependency_links = [
    'https://github.com/jeffh/describe/archive/907b42e4947f88111667a39e23bc5d5e0bf167fd.tar.gz#egg=describe',
]

install_requires = [
    'nexus>=0.3.0',
    'gutter>=0.1.1',
    'django',
    'ordereddict',
]

setup_requires = []
if 'nosetests' in sys.argv[1:]:
    setup_requires.append('nose')


setup(
    name='gutter-django',
    version='0.1.9',
    author='DISQUS',
    author_email='opensource@disqus.com',
    url='http://github.com/disqus/gutter-django',
    description='Web UI to administer Gutter switches.',
    packages=find_packages(exclude=["example_project", "tests"]),
    zip_safe=False,
    install_requires=install_requires,
    setup_requires=setup_requires,
    license='Apache License 2.0',
    tests_require=tests_require,
    extras_require={'test': tests_require},
    namespace_packages=['gutter'],
    test_suite='nose.collector',
    include_package_data=True,
    classifiers=[
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Operating System :: OS Independent',
        'Topic :: Software Development'
    ],
)
